@@ -1,5 +1,6 @@
 from blatann.nrf.nrf_types import BLEHci as _BLEHci
-from blatann.gap.smp import SecurityStatus, IoCapabilities, AuthenticationKeyType, SecurityParameters
+from blatann.gap.smp import (SecurityStatus, IoCapabilities, AuthenticationKeyType,
+                             SecurityParameters, PairingPolicy, SecurityLevel)
 from blatann.gap.scanning import ScanParameters
 from blatann.gap.advertising import AdvertisingData, AdvertisingFlags
 
