@@ -1,3 +1,4 @@
+from __future__ import annotations
 import logging
 from blatann.nrf import nrf_events, nrf_types
 from blatann import exceptions
@@ -16,9 +17,14 @@ MAX_ADVERTISING_INTERVAL_MS = nrf_types.adv_interval_range.max
 
 
 class Advertiser(object):
+    """
+    Class which manages the advertising state of the BLE Device
+    """
     # Constant used to indicate that the BLE device should advertise indefinitely, until
     # connected to or stopped manually
     ADVERTISE_FOREVER = 0
+    """Special value used to indicate that the BLE device should advertise indefinitely 
+       until either a central is connected or stopped manually."""
 
     def __init__(self, ble_device, client, conn_tag=0):
         """
@@ -39,27 +45,30 @@ class Advertiser(object):
         self._conn_tag = conn_tag
 
     @property
-    def on_advertising_timeout(self):
+    def on_advertising_timeout(self) -> Event[Advertiser, None]:
         """
         Event generated whenever advertising times out and finishes with no connections made
-        Event args: None
+
+        .. note:: If auto-restart advertising is enabled, this will trigger on each advertising timeout configured
 
         :return: an Event which can have handlers registered to and deregistered from
-        :rtype: Event
         """
         return self._on_advertising_timeout
 
     @property
-    def is_advertising(self):
+    def is_advertising(self) -> bool:
         """
+        **Read Only**
+
         Current state of advertising
-        :return:
         """
         return self._is_advertising
 
     @property
     def min_interval_ms(self) -> float:
         """
+        **Read Only**
+
         The minimum allowed advertising interval, in millseconds.
         This is defined by the Bluetooth specification.
         """
@@ -68,25 +77,47 @@ class Advertiser(object):
     @property
     def max_interval_ms(self) -> float:
         """
+        **Read Only**
+
         The maximum allowed advertising interval, in milliseconds.
         This is defined by the Bluetooth specification.
         """
         return MAX_ADVERTISING_INTERVAL_MS
 
-    def set_advertise_data(self, advertise_data=AdvertisingData(), scan_response=AdvertisingData()):
+    @property
+    def auto_restart(self) -> bool:
+        """
+        Enables/disables whether or not the device should automatically restart
+        advertising when an advertising timeout occurs or the client is disconnected.
+
+        .. note:: Auto-restart is disabled automatically when :meth:`stop` is called
+
+        :getter: Gets the auto-restart flag
+        :setter: Sets/clears the auto-restart flag
+        """
+        return self._auto_restart
+
+    @auto_restart.setter
+    def auto_restart(self, value: bool):
+        self._auto_restart = bool(value)
+
+    def set_advertise_data(self,
+                           advertise_data: AdvertisingData = AdvertisingData(),
+                           scan_response: AdvertisingData = AdvertisingData()):
         """
         Sets the advertising and scan response data which will be broadcasted to peers during advertising
 
-        Note: BLE Restricts advertise and scan response data to an encoded length of 31 bytes each.
-        Use AdvertisingData.check_encoded_length() to determine if the
+        .. note:: BLE Restricts advertise and scan response data to an encoded length of 31 bytes each.
+           Use :meth:`AdvertisingData.check_encoded_length() <blatann.gap.advertise_data.AdvertiseData.check_encoded_length>`
+           to determine if the payload is too large
 
-        :param advertise_data: The advertise data to use
-        :type advertise_data: AdvertisingData
-        :param scan_response: The scan response data to use
-        :type scan_response: AdvertisingData
+        :param advertise_data: The advertising data to use
+        :param scan_response: The scan response data to use.
+                              This data is only sent when a scanning device requests the scan response packet (active scanning)
+        :raises: InvalidOperationException if one of the payloads is too large
         """
         adv_len, adv_pass = advertise_data.check_encoded_length()
-        scan_len, scan_pass = advertise_data.check_encoded_length()
+        scan_len, scan_pass = scan_response.check_encoded_length()
 
         if not adv_pass:
             raise exceptions.InvalidOperationException("Encoded Advertising data length is too long ({} bytes). "
@@ -98,7 +129,10 @@ class Advertiser(object):
 
         self.ble_device.ble_driver.ble_gap_adv_data_set(advertise_data.to_ble_adv_data(), scan_response.to_ble_adv_data())
 
-    def set_default_advertise_params(self, advertise_interval_ms, timeout_seconds, advertise_mode=AdvertisingMode.connectable_undirected):
+    def set_default_advertise_params(self,
+                                     advertise_interval_ms: float,
+                                     timeout_seconds: int,
+                                     advertise_mode: AdvertisingMode = AdvertisingMode.connectable_undirected):
         """
         Sets the default advertising parameters so they do not need to be specified on each start
 
@@ -106,25 +140,28 @@ class Advertiser(object):
                                       Should be a multiple of 0.625ms, otherwise it'll be rounded down to the nearest 0.625ms
         :param timeout_seconds: How long to advertise for before timing out, in seconds. For no timeout, use ADVERTISE_FOREVER (0)
         :param advertise_mode: The mode the advertiser should use
-        :type advertise_mode: AdvertisingMode
         """
         nrf_types.adv_interval_range.validate(advertise_interval_ms)
         self._advertise_interval = advertise_interval_ms
         self._timeout = timeout_seconds
         self._advertise_mode = advertise_mode
 
-    def start(self, adv_interval_ms=None, timeout_sec=None, auto_restart=None, advertise_mode=None):
+    def start(self,
+              adv_interval_ms: float = None,
+              timeout_sec: int = None,
+              auto_restart: bool = None,
+              advertise_mode: AdvertisingMode = None):
         """
-        Starts advertising with the given parameters. If none given, will use the default
+        Starts advertising with the given parameters. If none given, will use the default set through :meth:`set_default_advertise_params`
 
         :param adv_interval_ms: The interval at which to send out advertise packets, in milliseconds.
-                                Should be a multiple of 0.625ms, otherwise it'll be rounde down to the nearest 0.625ms
+                                Should be a multiple of 0.625ms, otherwise it'll be round down to the nearest 0.625ms
         :param timeout_sec: The duration which to advertise for. For no timeout, use ADVERTISE_FOREVER (0)
         :param auto_restart: Flag indicating that advertising should restart automatically when the timeout expires, or
                              when the client disconnects
         :param advertise_mode: The mode the advertiser should use
-        :return: A waitable that will expire either when the timeout occurs, or a client connects.
-                 Waitable Returns a peer.Client() object
+        :return: A waitable that will expire either when the timeout occurs or a client connects.
+                 The waitable will return either ``None`` on timeout or :class:`~blatann.peer.Client` on successful connection
         :rtype: ClientConnectionWaitable
         """
         if self._is_advertising:
@@ -174,10 +211,12 @@ class Advertiser(object):
         :type event: nrf_events.GapEvtTimeout
         """
         if event.src == nrf_events.BLEGapTimeoutSrc.advertising:
-            self._is_advertising = False
+            # Notify that advertising timed out first which may call stop() to disable auto-restart
             self._on_advertising_timeout.notify(self)
             if self._auto_restart:
                 self._start()
+            else:
+                self._is_advertising = False
 
     def _handle_connect(self, peer, event):
         self._is_advertising = False
