@@ -61,7 +61,7 @@ def NordicSemiErrorCheck(wrapped=None, expected=driver.NRF_SUCCESS):
 
     @wrapt.decorator
     def wrapper(wrapped, instance, args, kwargs):
-        logger.debug("[{}] {}{}".format(instance.serial_port, wrapped.__name__, args))
+        logger.debug("[%s] %s%s", instance.serial_port, wrapped.__name__, args)
         result = wrapped(*args, **kwargs)
         if isinstance(result, (list, tuple)):
             err_code = result[0]
@@ -77,7 +77,7 @@ def NordicSemiErrorCheck(wrapped=None, expected=driver.NRF_SUCCESS):
                 err_string = 'Error code: {}'.format(NrfError(err_code))
             except ValueError:
                 err_string = 'Error code: 0x{:04x}, {}'.format(err_code, err_code)
-            raise NordicSemiException('Failed to {}. {}'.format(wrapped.__name__, err_string))
+            raise NordicSemiException('Failed to {}. {}'.format(wrapped.__name__, err_string), err_code)
         return result
 
     return wrapper(wrapped)
@@ -89,7 +89,6 @@ class NrfDriverObserver(object):
 
 
 class NrfDriver(object):
-    api_lock = Lock()
     default_baud_rate = 1000000
     ATT_MTU_DEFAULT = driver.BLE_GATT_ATT_MTU_DEFAULT
 
@@ -121,7 +120,7 @@ class NrfDriver(object):
         return self._serial_port
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def open(self):
         if self.is_open:
             logger.warning("Trying to open already opened driver")
@@ -156,7 +155,7 @@ class NrfDriver(object):
         return self._event_thread is not None
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def close(self):
         if not self.is_open:
             return driver.NRF_SUCCESS
@@ -228,7 +227,7 @@ class NrfDriver(object):
     BLE Generic methods
     """
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_conn_configure(self, conn_params):
         assert isinstance(conn_params, BleConnConfig)
         for tag, cfg in conn_params.get_configs():
@@ -238,7 +237,7 @@ class NrfDriver(object):
         return driver.NRF_SUCCESS
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_enable(self, ble_enable_params=None):
         if not ble_enable_params:
             ble_enable_params = self.ble_enable_params_setup()
@@ -252,12 +251,24 @@ class NrfDriver(object):
         return driver.sd_ble_enable(self.rpc_adapter, None)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
+    def ble_opt_set(self, ble_opt):
+        assert isinstance(ble_opt, BleOption)
+        opt = driver.ble_opt_t()
+        sub_attr = opt
+        path = ble_opt.path.split(".")
+        for name in path[:-1]:
+            sub_attr = getattr(sub_attr, name)
+        setattr(sub_attr, path[-1], ble_opt.to_c())
+        return driver.sd_ble_opt_set(self.rpc_adapter, ble_opt.option_flag, opt)
+
+    @NordicSemiErrorCheck
+    @wrapt.synchronized
     def ble_user_mem_reply(self, conn_handle):
         return driver.sd_ble_user_mem_reply(self.rpc_adapter, conn_handle, None)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_vs_uuid_add(self, uuid_base):
         assert isinstance(uuid_base, BLEUUIDBase), 'Invalid argument type'
         uuid_type = driver.new_uint8()
@@ -274,7 +285,7 @@ class NrfDriver(object):
     """
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_addr_get(self):
         addr = driver.ble_gap_addr_t()
         err_code = driver.sd_ble_gap_addr_get(self.rpc_adapter, addr)
@@ -285,14 +296,36 @@ class NrfDriver(object):
         return err_code, addr
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_addr_set(self, address):
         assert isinstance(address, BLEGapAddr), "Invalid argument type"
         addr_c = address.to_c()
         return driver.sd_ble_gap_addr_set(self.rpc_adapter, addr_c)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
+    def ble_gap_device_name_set(self, name):
+        if isinstance(name, str):
+            name = name.encode("utf8")
+        write_perm = BLEGapSecMode(0, 0).to_c()
+        length = len(name)
+        data = util.list_to_uint8_array(name).cast()
+        return driver.sd_ble_gap_device_name_set(self.rpc_adapter, write_perm, data, length)
+
+    @NordicSemiErrorCheck
+    @wrapt.synchronized
+    def ble_gap_appearance_set(self, value):
+        return driver.sd_ble_gap_appearance_set(self.rpc_adapter, value)
+
+    @NordicSemiErrorCheck
+    @wrapt.synchronized
+    def ble_gap_ppcp_set(self, conn_params):
+        assert isinstance(conn_params, BLEGapConnParams), "Invalid argument type"
+        params = conn_params.to_c()
+        return driver.sd_ble_gap_ppcp_set(self.rpc_adapter, params)
+
+    @NordicSemiErrorCheck
+    @wrapt.synchronized
     def ble_gap_adv_start(self, adv_params=None, conn_cfg_tag=0):
         if not adv_params:
             adv_params = self.adv_params_setup()
@@ -300,7 +333,7 @@ class NrfDriver(object):
         return driver.sd_ble_gap_adv_start(self.rpc_adapter, adv_params.to_c(), conn_cfg_tag)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_conn_param_update(self, conn_handle, conn_params):
         assert isinstance(conn_params, (BLEGapConnParams, NoneType)), 'Invalid argument type'
         if conn_params:
@@ -308,12 +341,12 @@ class NrfDriver(object):
         return driver.sd_ble_gap_conn_param_update(self.rpc_adapter, conn_handle, conn_params)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_adv_stop(self):
         return driver.sd_ble_gap_adv_stop(self.rpc_adapter)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_scan_start(self, scan_params=None):
         if not scan_params:
             scan_params = self.scan_params_setup()
@@ -321,12 +354,12 @@ class NrfDriver(object):
         return driver.sd_ble_gap_scan_start(self.rpc_adapter, scan_params.to_c())
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_scan_stop(self):
         return driver.sd_ble_gap_scan_stop(self.rpc_adapter)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_connect(self, address, scan_params=None, conn_params=None, conn_cfg_tag=0):
         assert isinstance(address, BLEGapAddr), 'Invalid argument type'
 
@@ -345,7 +378,7 @@ class NrfDriver(object):
                                          conn_cfg_tag)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_disconnect(self, conn_handle, hci_status_code=BLEHci.remote_user_terminated_connection):
         assert isinstance(hci_status_code, BLEHci), 'Invalid argument type'
         return driver.sd_ble_gap_disconnect(self.rpc_adapter,
@@ -353,7 +386,7 @@ class NrfDriver(object):
                                             hci_status_code.value)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_adv_data_set(self, adv_data=BLEAdvData(), scan_data=BLEAdvData()):
         assert isinstance(adv_data, BLEAdvData), 'Invalid argument type'
         assert isinstance(scan_data, BLEAdvData), 'Invalid argument type'
@@ -367,7 +400,7 @@ class NrfDriver(object):
                                               scan_data_len)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_data_length_update(self, conn_handle, params=None):
         assert isinstance(params, (BLEGapDataLengthParams, NoneType))
         if isinstance(params, BLEGapDataLengthParams):
@@ -375,7 +408,7 @@ class NrfDriver(object):
         return driver.sd_ble_gap_data_length_update(self.rpc_adapter, conn_handle, params, None)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_phy_update(self, conn_handle, tx_phy=BLEGapPhy.auto, rx_phy=BLEGapPhy.auto):
         params = BLEGapPhys(tx_phy, rx_phy)
         return driver.sd_ble_gap_phy_update(self.rpc_adapter, conn_handle, params.to_c())
@@ -385,7 +418,7 @@ class NrfDriver(object):
     """
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_authenticate(self, conn_handle, sec_params):
         assert isinstance(sec_params, (BLEGapSecParams, NoneType)), 'Invalid argument type'
         return driver.sd_ble_gap_authenticate(self.rpc_adapter,
@@ -393,7 +426,7 @@ class NrfDriver(object):
                                               sec_params.to_c() if sec_params else None)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_sec_params_reply(self, conn_handle, sec_status, sec_params, sec_keyset):
         assert isinstance(sec_status, BLEGapSecStatus), 'Invalid argument type'
         assert isinstance(sec_params, (BLEGapSecParams, NoneType)), 'Invalid argument type'
@@ -406,7 +439,7 @@ class NrfDriver(object):
                                                   sec_keyset.to_c())
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_auth_key_reply(self, conn_handle, key_type, key):
         if key is not None:
             key_buf = util.list_to_uint8_array(key).cast()
@@ -415,7 +448,7 @@ class NrfDriver(object):
         return driver.sd_ble_gap_auth_key_reply(self.rpc_adapter, conn_handle, key_type, key_buf)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_sec_info_reply(self, conn_handle, enc_info=None, irk=None, sign_info=None):
         assert isinstance(enc_info, (BLEGapEncryptInfo, NoneType)), "Invalid argument type"
         assert isinstance(irk, (BLEGapIdKey, NoneType)), "Invalid argument type"
@@ -430,32 +463,15 @@ class NrfDriver(object):
         return driver.sd_ble_gap_sec_info_reply(self.rpc_adapter, conn_handle, enc_info, irk, sign_info)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
-    def ble_gap_encrypt(self, conn_handle, ediv, rand, ltk, lesc, auth):
-        # TODO: Clean up
-        # assert isinstance(sec_params, (BLEGapSecParams, NoneType)), 'Invalid argument type'
-        # assert isinstance(sec_keyset, BLEGapSecKeyset), 'Invalid argument type'
-        # print 'ediv %r' % master_id.ediv
-        # print 'rand %r' % util.uint8_array_to_list(master_id.rand, 8)
-        # print 'ltk  %r' % util.uint8_array_to_list(enc_info.ltk, enc_info.ltk_len)
-        # print 'len  %r' % enc_info.ltk_len
-        # print 'lesc %r' % enc_info.lesc
-        # print 'auth %r' % enc_info.auth
+    @wrapt.synchronized
+    def ble_gap_encrypt(self, conn_handle, master_id, enc_info):
+        assert isinstance(master_id, BLEGapMasterId)
+        assert isinstance(enc_info, BLEGapEncryptInfo)
 
-        rand_arr = util.list_to_uint8_array(rand)
-        ltk_arr = util.list_to_uint8_array(ltk)
-        master_id = driver.ble_gap_master_id_t()
-        master_id.ediv = ediv
-        master_id.rand = rand_arr.cast()
-        enc_info = driver.ble_gap_enc_info_t()
-        enc_info.ltk_len = len(ltk)
-        enc_info.ltk = ltk_arr.cast()
-        enc_info.lesc = lesc
-        enc_info.auth = auth
-        return driver.sd_ble_gap_encrypt(self.rpc_adapter, conn_handle, master_id, enc_info)
+        return driver.sd_ble_gap_encrypt(self.rpc_adapter, conn_handle, master_id.to_c(), enc_info.to_c())
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gap_lesc_dhkey_reply(self, conn_handle, dh_key):
         assert isinstance(dh_key, BLEGapDhKey)
 
@@ -469,7 +485,7 @@ class NrfDriver(object):
     # sd_ble_gatts_initial_user_handle_get, sd_ble_gatts_attr_get
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gatts_service_add(self, service_type, uuid, service_handle):
         handle = driver.new_uint16()
         uuid_c = uuid.to_c()
@@ -482,7 +498,7 @@ class NrfDriver(object):
         return err_code
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gatts_characteristic_add(self, service_handle, char_md, attr_char_value, char_handle):
         # TODO type assertions
         handle_params = driver.ble_gatts_char_handles_t()
@@ -499,17 +515,28 @@ class NrfDriver(object):
         return err_code
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
+    def ble_gatts_descriptor_add(self, char_handle, attr):
+        assert isinstance(attr, BLEGattsAttribute)
+        handle = driver.new_uint16()
+        err_code = driver.sd_ble_gatts_descriptor_add(self.rpc_adapter, char_handle, attr.to_c(), handle)
+        if err_code == driver.NRF_SUCCESS:
+            attr.handle = driver.uint16_value(handle)
+        return err_code
+
+
+    @NordicSemiErrorCheck
+    @wrapt.synchronized
     def ble_gatts_rw_authorize_reply(self, conn_handle, authorize_reply_params):
         assert isinstance(authorize_reply_params, BLEGattsRwAuthorizeReplyParams)
         return driver.sd_ble_gatts_rw_authorize_reply(self.rpc_adapter, conn_handle, authorize_reply_params.to_c())
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
-    def ble_gatts_value_get(self, conn_handle, attribute_handle, gatts_value):
+    @wrapt.synchronized
+    def ble_gatts_value_get(self, conn_handle, attribute_handle, gatts_value, max_bytes_read=512):
         assert isinstance(gatts_value, BLEGattsValue)
         value_params = gatts_value.to_c()
-        value_params.len = 512  # Allow up to 512 bytes to be read
+        value_params.len = max_bytes_read
         err_code = driver.sd_ble_gatts_value_get(self.rpc_adapter, conn_handle, attribute_handle, value_params)
 
         if err_code == driver.NRF_SUCCESS:
@@ -519,34 +546,45 @@ class NrfDriver(object):
         return err_code
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gatts_value_set(self, conn_handle, attribute_handle, gatts_value):
         assert isinstance(gatts_value, BLEGattsValue)
         value_params = gatts_value.to_c()
         return driver.sd_ble_gatts_value_set(self.rpc_adapter, conn_handle, attribute_handle, value_params)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gatts_hvx(self, conn_handle, hvx_params):
         assert isinstance(hvx_params, BLEGattsHvx)
         return driver.sd_ble_gatts_hvx(self.rpc_adapter, conn_handle, hvx_params.to_c())
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gatts_service_changed(self, conn_handle, start_handle, end_handle):
         return driver.sd_ble_gatts_service_changed(self.rpc_adapter, conn_handle, start_handle, end_handle)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gatts_exchange_mtu_reply(self, conn_handle, server_mtu):
         return driver.sd_ble_gatts_exchange_mtu_reply(self.rpc_adapter, conn_handle, server_mtu)
 
+    @NordicSemiErrorCheck
+    @wrapt.synchronized
+    def ble_gatts_sys_attr_set(self, conn_handle, sys_attr_data, flags=0):
+        if sys_attr_data is not None:
+            data = util.list_to_uint8_array(sys_attr_data).cast()
+            length = len(sys_attr_data)
+        else:
+            data = None
+            length = 0
+        return driver.sd_ble_gatts_sys_attr_set(self.rpc_adapter, conn_handle, data, length, flags)
+
     """
     GATTC Methods
     """
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gattc_write(self, conn_handle, write_params):
         assert isinstance(write_params, BLEGattcWriteParams), 'Invalid argument type'
         return driver.sd_ble_gattc_write(self.rpc_adapter,
@@ -554,7 +592,7 @@ class NrfDriver(object):
                                          write_params.to_c())
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gattc_prim_srvc_disc(self, conn_handle, srvc_uuid, start_handle):
         assert isinstance(srvc_uuid, (BLEUUID, NoneType)), 'Invalid argument type'
         return driver.sd_ble_gattc_primary_services_discover(self.rpc_adapter,
@@ -563,7 +601,7 @@ class NrfDriver(object):
                                                              srvc_uuid.to_c() if srvc_uuid else None)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gattc_char_disc(self, conn_handle, start_handle, end_handle):
         handle_range = driver.ble_gattc_handle_range_t()
         handle_range.start_handle = start_handle
@@ -573,7 +611,7 @@ class NrfDriver(object):
                                                             handle_range)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gattc_desc_disc(self, conn_handle, start_handle, end_handle):
         handle_range = driver.ble_gattc_handle_range_t()
         handle_range.start_handle = start_handle
@@ -583,7 +621,7 @@ class NrfDriver(object):
                                                         handle_range)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gattc_attr_info_disc(self, conn_handle, start_handle, end_handle):
         handle_range = driver.ble_gattc_handle_range_t()
         handle_range.start_handle = start_handle
@@ -591,19 +629,19 @@ class NrfDriver(object):
         return driver.sd_ble_gattc_attr_info_discover(self.rpc_adapter, conn_handle, handle_range)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gattc_read(self, conn_handle, read_handle, offset=0):
         return driver.sd_ble_gattc_read(self.rpc_adapter, conn_handle, read_handle, offset)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gattc_exchange_mtu_req(self, conn_handle, att_mtu_size):
         return driver.sd_ble_gattc_exchange_mtu_request(self.rpc_adapter,
                                                         conn_handle,
                                                         att_mtu_size)
 
     @NordicSemiErrorCheck
-    @wrapt.synchronized(api_lock)
+    @wrapt.synchronized
     def ble_gattc_hv_confirm(self, conn_handle, attr_handle):
         return driver.sd_ble_gattc_hv_confirm(self.rpc_adapter, conn_handle, attr_handle)
 
@@ -612,8 +650,8 @@ class NrfDriver(object):
     """
 
     def _status_handler(self, adapter, status_code, status_message):
-        # print(status_message)
-        pass
+        if self._log_driver_comms:
+            logger.info("STATUS [{}]: {}".format(status_code, status_message))
 
     def _log_message_handler(self, adapter, severity, log_message):
         if self._log_driver_comms:
